from typing import Callable, List

from neurosym.dsl.dsl import ROOT_SYMBOL
from neurosym.types.type_with_environment import Environment, TypeWithEnvironment

from .preorder_mask import PreorderMask


class TypePreorderMask(PreorderMask):
    """
    Masks out productions that would lead to ill-typed programs. This mask is
    computed by traversing the tree in a preorder fashion and checking the types
    of the children of the current node.

    This mask can be cached by the tree distribution to speed up enumeration, the
    cache key is the unique hash of the ``TypeWithEnvironment`` of the current node.
    """

    def __init__(self, tree_dist, dsl):
        super().__init__(tree_dist)
        self.dsl = dsl
        assert len(self.dsl.valid_root_types) == 1, "Only one root type is supported"
        self.root_type = self.dsl.valid_root_types[0]

        # stack of list of type_with_environment objects
        # each list represents the types of the children of the current node
        self.type_stack: List[List[TypeWithEnvironment]] = []

    def compute_mask(self, position, symbols):
        valid_productions = {
            self.tree_dist.symbol_to_index[sym.symbol()]
            for sym, _ in self.dsl.productions_for_type(self.type_stack[-1][position])
        }
        return [i in valid_productions for i in symbols]

    def on_entry(self, position, symbol) -> Callable[[], None]:
        symbol, arity = self.tree_dist.symbols[symbol]
        if symbol == ROOT_SYMBOL:
            self.type_stack.append(
                [TypeWithEnvironment(self.root_type, Environment.empty())]
            )
            return self.type_stack.pop
        parent_type = self.type_stack[-1][position]
        production = self.dsl.get_production(symbol)
        children_types = production.type_signature().unify_return(parent_type)
        if children_types is None:
            raise ValueError(
                f"Type mismatch in production {production} with parent type {parent_type}"
            )
        assert len(children_types) == arity
        self.type_stack.append(children_types)
        return self.type_stack.pop

    def on_exit(self, position, symbol) -> Callable[[], None]:
        del position
        del symbol
        last = self.type_stack.pop()
        return lambda: self.type_stack.append(last)

    def cache_key(self, parents):
        position = parents[-1][1]
        typ = self.type_stack[-1][position]
        return typ.unique_hash



import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import Callable, Dict, Iterator, List, Tuple

import numpy as np


class Type(ABC):
    def node_summary(self):
        """
        Return a summary of this node, excluding children. This is useful
        for caching in certain contexts, such as the TreeTrie.
        """
        summary_dict = {
            "_type": self.__class__.__name__,
            **self.inherent_parameters(),
        }
        summary_dict = dict(sorted(summary_dict.items()))
        return str(summary_dict)

    @abstractmethod
    def inherent_parameters(self):
        """
        Return a list of inherent parameters of the type, i.e., all parameters not associated
        with the children.
        """
        raise NotImplementedError

    @abstractmethod
    def children(self):
        """
        Iterate over direct children and yield each.
        """
        raise NotImplementedError

    @abstractmethod
    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        """
        Unify this type with another type. Returns a dictionary of substitutions.

        :param other: the other type to unify with.
        :param already_tried_other_direction: then we have already tried to unify
            the other direction and failed, so we should not try again.

        :raises UnificationError: if the types cannot be unified.
        """
        raise NotImplementedError

    @abstractmethod
    def subst_type_vars(self, subst: Dict[str, "Type"]):
        """
        Substitute type variables in the type according to the substitution.
        """
        raise NotImplementedError

    def is_atomic(self):
        """
        Return True if the type is atomic.
        """
        return len(list(self.walk_type_nodes())) == 1 and not isinstance(
            self, TypeVariable
        )

    def depth_of_type_variables(self) -> List[Tuple[str, float]]:
        """
        Return a list of tuples of the form (type_variable, depth) where depth is the
        depth of the type variable in the type tree.
        """
        node_height = np.log2(1 + len(list(self.children())))
        return [
            (tv, depth + node_height)
            for child in self.children()
            for tv, depth in child.depth_of_type_variables()
        ]

    def max_depth_per_type_variable(self) -> Dict[str, float]:
        """
        Return a dictionary mapping type variables to their maximum depth in the type
        tree.
        """
        result = {}
        for tv, depth in self.depth_of_type_variables():
            result[tv] = max(result.get(tv, 0), depth)
        return result

    def get_type_vars(self):
        """
        Return a set of all type variables in the type.
        """
        return sorted({tv for tv, _ in self.depth_of_type_variables()})

    def has_type_vars(self):
        """
        Return True if the type has type variables.
        """
        return len(self.get_type_vars()) > 0

    @cached_property
    def depth(self):
        """
        Return the depth of the type tree.
        """
        children = [child.depth for child in self.children()]
        depth = max(children + [0]) + np.log2(len(children) + 1)
        return depth

    def walk_type_nodes(self) -> Iterator["Type"]:
        """
        Walk the type tree and yield all children nodes.
        """
        yield self
        for child in self.children():
            yield from child.walk_type_nodes()


class UnificationError(Exception):
    """
    Raised when two types cannot be unified.
    """


@dataclass(frozen=True, eq=True)
class AtomicType(Type):
    """
    An atomic type is a type that cannot be further decomposed. It is a leaf node in the
    type tree. Examples of atomic types are integers, floats, and strings. These types
    are rendered as strings of their name, e.g., ``AtomicType("f")`` is rendered as ``"f"``.

    :field name: the name of the atomic type.
    """

    name: str

    def __post_init__(self):
        assert self.name.isidentifier(), f"{self.name} is not a valid identifier"

    def inherent_parameters(self):
        return dict(name=self.name)

    def children(self):
        yield from []

    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        if isinstance(other, AtomicType):
            if self.name != other.name:
                raise UnificationError(f"{self} != {other}")
            return {}
        if already_tried_other_direction:
            raise UnificationError(f"{self} != {other}")
        return other.unify(self, already_tried_other_direction=True)

    def subst_type_vars(self, subst: Dict[str, Type]):
        return self


@dataclass(frozen=True, eq=True)
class TensorType(Type):
    """
    A tensor is a type that represents a tensor of a given shape. These types are rendered
    as strings of the form ``{dtype, *shape}``, e.g., ``TensorType(ns.AtomicType("f"), (3, 4))`` is
    rendered as ``"{f, 3, 4}"``.

    :field dtype: the data type of the tensor (usually ``ns.AtomicType("f")``).
    :field shape: the shape of the tensor.
    """

    dtype: Type
    shape: Tuple[int]

    def inherent_parameters(self):
        return dict(dtype=self.dtype, shape=self.shape)

    def children(self):
        yield from []

    def subst_type_vars(self, subst: Dict[str, Type]):
        del subst
        assert not self.has_type_vars()
        return self

    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        if isinstance(other, TensorType):
            if self.shape != other.shape:
                raise UnificationError(f"{self} != {other}")
            return self.dtype.unify(other.dtype)
        if already_tried_other_direction:
            raise UnificationError(f"{self} != {other}")
        return other.unify(self, already_tried_other_direction=True)


@dataclass(frozen=True, eq=True)
class ListType(Type):
    """
    Represents a list of elements of a given type. These types are rendered as strings of
    the form ``[element_type]``, e.g., ``ListType(ns.AtomicType("f"))`` is rendered as ``"[f]"``.

    :field element_type: the type of the elements in the list.
    """

    element_type: Type

    def inherent_parameters(self):
        return dict()

    def children(self):
        yield self.element_type

    def subst_type_vars(self, subst: Dict[str, Type]):
        return ListType(self.element_type.subst_type_vars(subst))

    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        if isinstance(other, ListType):
            return self.element_type.unify(other.element_type)
        if already_tried_other_direction:
            raise UnificationError(f"{self} != {other}")
        return other.unify(self, already_tried_other_direction=True)


@dataclass(frozen=True, eq=True)
class ArrowType(Type):
    """
    An arrow type represents a function type. It is a type that takes a tuple of input
    types and returns an output type. These types are rendered as strings of the form
    ``(input_type1, input_type2, ...) -> (output_type)``, e.g.,
    ``ArrowType((ns.AtomicType("f"), ns.AtomicType("f")), ns.AtomicType("f"))`` is rendered
    as ``"(f, f) -> f"``.

    :field input_type: a tuple of input types.
    :field output_type: the output type.
    """

    input_type: Tuple[Type]
    output_type: Type

    def __post_init__(self):
        assert isinstance(self.input_type, tuple), "input_type must be a tuple"

    def subst_type_vars(self, subst: Dict[str, Type]):
        return ArrowType(
            tuple(t.subst_type_vars(subst) for t in self.input_type),
            self.output_type.subst_type_vars(subst),
        )

    def inherent_parameters(self):
        return dict()

    def children(self):
        yield from self.input_type
        yield self.output_type

    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        if not isinstance(other, ArrowType):
            if already_tried_other_direction:
                raise UnificationError(f"{self} != {other}")
            return other.unify(self, already_tried_other_direction=True)
        if len(self.input_type) != len(other.input_type):
            raise UnificationError(f"{self} != {other}")
        individuals = []
        for t1, t2 in zip(self.input_type, other.input_type):
            individuals.append(t1.unify(t2))
        individuals.append(self.output_type.unify(other.output_type))
        subst = {}
        for individual in individuals:
            for k, v in individual.items():
                if k in subst:
                    if subst[k] != v:
                        raise UnificationError(f"{k} is unified to {subst[k]} and {v}")
                else:
                    subst[k] = v
        return subst


@dataclass(frozen=True, eq=True)
class GenericTypeVariable(Type):
    """
    Base class for type variables. See ``TypeVariable`` and ``FilteredTypeVariable`` for
    concrete implementations and documentation
    """

    name: str

    @classmethod
    def fresh(cls):
        return cls(f"_{uuid.uuid4().hex}")

    def __post_init__(self):
        assert self.name.isidentifier(), f"{self.name} is not a valid identifier"

    def depth_of_type_variables(self) -> List[Tuple[str, int]]:
        return [(self.name, 0)]

    def subst_type_vars(self, subst: Dict[str, Type]):
        return subst.get(self.name, self)

    def inherent_parameters(self):
        return dict()

    def children(self):
        yield from []

    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        return {self.name: other}


class TypeVariable(GenericTypeVariable):
    """
    A type variable is a type that can be unified with any other type, but must be
    unified with the same type throughout the type tree. Type variables are used to
    represent generic types in the type system. These types are rendered as strings of
    the form ``#name``, e.g., ``TypeVariable("x")`` is rendered as ``#x``.

    :field name: the name of the type variable.
    """


@dataclass(frozen=True, eq=True)
class FilteredTypeVariable(GenericTypeVariable):
    """
    A filtered type variable is a type variable (see ``TypeVariable``) that can only be
    unified with types that satisfy a given predicate. These types are rendered as
    strings of the form ``%name``, e.g., ``FilteredTypeVariable("x", lambda t: isinstance(t, ns.AtomicType))``
    is rendered as ``%x``.

    :field name: the name of the type variable.
    :field type_filter: a predicate that takes a type and returns True if the type can be unifie
        with the type variable.
    """

    type_filter: Callable[[Type], bool]

    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        if self.type_filter(other):
            return {self.name: other}
        raise UnificationError(f"{self} does not match {other}")


float_t = AtomicType("f")



from typing import Callable, Tuple

from neurosym.program_dist.tree_distribution.preorder_mask.undos import (
    remove_last_n_elements,
)
from neurosym.python_dsl.names import PYTHON_DSL_SEPARATOR
from neurosym.utils.documentation import internal_only

from .handler import ConstructHandler, Handler


def create_target_handler(
    position: int, root_symbol: int, mask, defined_production_idxs, config
):
    """
    Create a target handler for the given root symbol. A target handler collects
    the assigned targets in the given expression.

    :param position: The position of the child in the parent.
    :param root_symbol: The root symbol of the expression.
    :param mask: The mask.
    :param defined_production_idxs: The defined production indices.
    :param config: The configuration.
    """
    targets_map = {
        "Name~L": NameTargetHandler,
        "arg~A": ArgTargetHandler,
        "alias~alias": AliasTargetHandler,
        "const-None~A": NonCollectingTargetHandler,
        "Subscript~L": NonCollectingTargetHandler,
        "Attribute~L": NonCollectingTargetHandler,
        "Tuple~L": TupleLHSHandler,
        "List~L": ListLHSHandler,
        "_starred_content~L": PassthroughLHSHandler,
        "_starred_starred~L": PassthroughLHSHandler,
        "Starred~L": StarredHandler,
        "arguments~As": ArgumentsHandler,
    }

    symbol = root_symbol
    symbol = mask.id_to_name(symbol)

    if symbol.startswith("const") and symbol.split(PYTHON_DSL_SEPARATOR)[-1] in {
        "Name",
        "NullableName",
        "NameStr",
        "NullableNameStr",
    }:
        return SymbolTargetHandler(mask, defined_production_idxs, config, root_symbol)

    pulled = config.pull_handler(
        position,
        symbol,
        mask,
        defined_production_idxs,
        handler_fn=create_target_handler,
    )
    if pulled is not None:
        return pulled

    if symbol.startswith("list"):
        return PassthroughLHSHandler(mask, defined_production_idxs, config)
    return targets_map[symbol](mask, defined_production_idxs, config)


class TargetHandler(Handler):
    """
    Handler that collects targets, stored in defined_symbols. This is used for
    LHS values that define symbols. It collects targets from any child frame
    by default.

    :field defined_symbols: The symbols defined by the construct.
    """

    def __init__(self, mask, defined_production_idxs, config):
        super().__init__(mask, defined_production_idxs, config)
        self.defined_symbols = []

    def on_child_exit(
        self, position: int, symbol: int, child: Handler
    ) -> Callable[[], None]:
        if hasattr(child, "defined_symbols"):
            symbols = [
                x for x in child.defined_symbols if x not in self.defined_symbols
            ]
            self.defined_symbols += symbols
            return remove_last_n_elements(self.defined_symbols, len(symbols))
        return lambda: None


@internal_only
class TargetConstructHandler(TargetHandler, ConstructHandler):
    def __init__(self, mask, defined_production_idxs, config):
        TargetHandler.__init__(self, mask, defined_production_idxs, config)
        ConstructHandler.__init__(self, mask, defined_production_idxs, config)


@internal_only
class PassthroughLHSHandler(TargetHandler):
    """
    Pass through handler that does not collect any information,
        instead it just targets the children at the given indices.

    If indices is None, it will target all children.
    """

    def on_child_enter(
        self, position: int, symbol: int
    ) -> Tuple[Handler, Callable[[], None]]:
        if self.is_defining(position):
            return self.target_child(position, symbol)
        return super().on_child_enter(position, symbol)

    def is_defining(self, position: int) -> bool:
        return True


@internal_only
class PassthroughLHSConstructHandler(PassthroughLHSHandler, TargetConstructHandler):
    """
    Pass through handler that is also a construct handler.
    """

    use_fields: list[str] = None

    def __init__(self, mask, defined_production_idxs, config):
        PassthroughLHSHandler.__init__(self, mask, defined_production_idxs, config)
        TargetConstructHandler.__init__(self, mask, defined_production_idxs, config)

        self.indices = [self.child_fields[x] for x in self.use_fields]

    def is_defining(self, position: int) -> bool:
        return self.indices is None or position in self.indices


@internal_only
class NonCollectingTargetHandler(PassthroughLHSHandler):
    """
    This is for LHS values where nothing is actually being defined
        (e.g., Subscript, Attribute, etc.)
    """

    def is_defining(self, position: int) -> bool:
        return False


@internal_only
class SymbolTargetHandler(TargetHandler):
    """
    Target handler for symbols. Since symbols have no children, this takes in the symbol
        and sets it up in the defined symbols in __init__.

    Do not keep the symbol if it is a const-None~NullableNameStr.
    """

    def __init__(self, mask, defined_production_idxs, config, symbol):
        super().__init__(mask, defined_production_idxs, config)
        if self.mask.id_to_name(symbol) != "const-None~NullableNameStr":
            self.defined_symbols = [symbol]

    def on_child_enter(
        self, position: int, symbol: int
    ) -> Tuple[Handler, Callable[[], None]]:
        raise NotImplementedError("symbols should not have children.")

    def on_child_exit(
        self, position: int, symbol: int, child: Handler
    ) -> Callable[[], None]:
        raise NotImplementedError("symbols should not have children.")

    def is_defining(self, position: int) -> bool:
        return True


@internal_only
class StarredHandler(PassthroughLHSConstructHandler):
    """
    Handles starred expressions.
    """

    name = "Starred~L"
    use_fields = ["value"]


@internal_only
class ArgumentsHandler(PassthroughLHSConstructHandler):
    """
    Handles arguments.
    """

    name = "arguments~As"
    use_fields = ("posonlyargs", "args", "vararg", "kwonlyargs", "kwarg")


@internal_only
class TupleLHSHandler(PassthroughLHSConstructHandler):
    """
    This is for LHS values where nothing is actually being defined
        (e.g., Subscript, Attribute, etc.)
    """

    name = "Tuple~L"
    use_fields = ["elts"]


@internal_only
class ListLHSHandler(TupleLHSHandler):
    """
    Handler for list LHS values.
    """

    name = "List~L"


@internal_only
class NameTargetHandler(TargetConstructHandler):
    """
    Handler for Name values, as well as other constructs containing symbols.
        Always keeps only the last defined symbol.
    """

    name = "Name~Name"
    # will select the last of these that is defined
    name_nodes = {"id"}

    def on_child_enter(
        self, position: int, symbol: int
    ) -> Tuple[Handler, Callable[[], None]]:
        if self.is_defining(position):
            return self.target_child(position, symbol)
        return super().on_child_enter(position, symbol)

    def on_child_exit(
        self, position: int, symbol: int, child: Handler
    ) -> Callable[[], None]:
        if hasattr(child, "defined_symbols") and child.defined_symbols:
            previous = self.defined_symbols
            self.defined_symbols = child.defined_symbols
            return lambda: setattr(self, "defined_symbols", previous)
        return lambda: None

    def is_defining(self, position: int) -> bool:
        return any(position == self.child_fields[x] for x in self.name_nodes)


@internal_only
class ArgTargetHandler(NameTargetHandler):
    name = "arg~arg"
    name_nodes = {"arg"}


@internal_only
class AliasTargetHandler(NameTargetHandler):
    name = "alias~alias"
    name_nodes = {"name", "asname"}



from dataclasses import dataclass
from typing import Dict, Tuple, Union


@dataclass(frozen=True, eq=True)
class SExpression:
    """
    Represents an s-expression, that is, an expression that contains a
    symbol and a list of children, which are also s-expressions or strings
    (leaves).

    :field symbol: The symbol of the s-expression.
    :field children: A tuple of children of the s-expression. Each child can
        be either a string or another s-expression.
    """

    symbol: str
    children: Tuple[Union["SExpression", str]]

    @property
    def postorder(self):
        for x in self.children:
            yield from x.postorder
        yield self

    def replace_symbols_by_id(self, id_to_symbol):
        """
        Replace symbols in the S-expression based on the given mapping. The current
        SExpression will not be mutated

        :param id_to_symbol: A mapping from id(node) to a new symbol.

        :return: A new S-expression with the symbols replaced.
        """
        return SExpression(
            id_to_symbol.get(id(self), self.symbol),
            tuple(child.replace_symbols_by_id(id_to_symbol) for child in self.children),
        )

    def replace_nodes_by_id(self, id_to_new_node):
        """
        Replace nodes in the S-expression based on the given mapping. Any remapped
        node will not have its children replaced. The current SExpression will
        not be mutated

        :param id_to_new_node: A mapping from id(node) to a new node.
        :return: A new S-expression with the nodes replaced.
        """
        if id(self) in id_to_new_node:
            return id_to_new_node[id(self)]
        return SExpression(
            self.symbol,
            tuple(child.replace_nodes_by_id(id_to_new_node) for child in self.children),
        )


@dataclass
class InitializedSExpression:
    symbol: str
    children: Tuple["InitializedSExpression"]
    # state includes things related to the execution of the program,
    # e.g. weights of a neural network
    state: Dict[str, object]

    def all_state_values(self):
        yield from self.state.values()
        for child in self.children:
            yield from child.all_state_values()



from abc import ABC, abstractmethod
from typing import Iterable, TypeVar

N = TypeVar("N")


class SearchGraph(ABC):
    """
    Represents a search graph where nodes are objects and edges are expansions of those objects.
    """

    @abstractmethod
    def initial_node(self) -> N:
        """
        The initial node of the search graph.
        """

    @abstractmethod
    def expand_node(self, node: N) -> Iterable[N]:
        """
        Find the neighbors of a node in the search graph.
        """

    @abstractmethod
    def is_goal_node(self, node: N) -> bool:
        """
        Return True iff the node is a goal node.
        """



import queue
from typing import Callable

from pathos.multiprocessing import ProcessingPool as Pool

from neurosym.programs.s_expression import SExpression
from neurosym.search.bounded_astar import BoundedAStarNode
from neurosym.search_graph.search_graph import SearchGraph


class FuturePriorityQueue(queue.PriorityQueue):
    """
    A priority queue with primitive support for futures.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.waiting = []

    def syncronize(self):
        # check if any of the waiting items are done
        for i, (item, item_fn) in enumerate(self.waiting):
            if item.ready():
                # add to the queue
                self.put(item_fn(item.get()))
                # remove from waiting
                self.waiting.pop(i)

    def putitem(self, item, future_fn: Callable):
        """Making sure this doesn't override the `put` method."""
        self.waiting.append((item, future_fn))
        self.syncronize()

    def empty(self):
        return len(self.waiting) == 0 and super().empty()

    def getitem(self):
        """Making sure this doesn't override the `get` method."""
        self.syncronize()
        return self.get_nowait()


def bounded_astar_async(
    g: SearchGraph,
    cost_plus_heuristic: Callable[[SExpression], float],
    max_depth: int,
    max_workers: int,
):
    """
    Performs a bounded a-star search on the given search graph, yielding each node in
    the order it was visited. Evaluates the cost_plus_heuristic function asynchronously
    spawning max_workers processes. This function doesn't optimally traverse the search
    graph.

    :param g: Search graph to search over
    :param cost_plus_heuristic: Cost plus heuristic function to use for A*.
        The heuristic function should be admissible, i.e. it should never overestimate
        the cost to reach the goal.
    :param max_depth: Maximum depth to search to.
    :param max_workers: Maximum number of workers to use for evaluating
        the cost_plus_heuristic function.
    """
    assert max_depth > 0, "Cannot have 0 depth."
    assert max_workers > 0, "Cannot have 0 workers."

    with Pool(max_workers) as executor:
        visited = set()
        fringe = FuturePriorityQueue()

        def add_to_fringe(node, depth):
            future = executor.apipe(cost_plus_heuristic, node)
            fringe.putitem(
                future,
                future_fn=lambda ret: BoundedAStarNode(ret, node, depth),
            )

        add_to_fringe(g.initial_node(), 0)
        while not fringe.empty():
            try:
                fringe_var = fringe.getitem()
                # pylint: disable=duplicate-code
                node, depth = fringe_var.node, fringe_var.depth
                if node.program in visited or depth > max_depth:
                    continue
                visited.add(node.program)
                if g.is_goal_node(node):
                    yield node
                for child in g.expand_node(node):
                    add_to_fringe(child, depth + 1)
            except queue.Empty:
                pass



import json
import os


def update_links(notebook_json, path):
    for cell in notebook_json["cells"]:
        if cell["cell_type"] == "markdown":
            cell["source"] = [
                x.replace(
                    path, path.replace("_solutions.ipynb", "_skeleton.ipynb")
                ).replace(" - Solution", "")
                for x in cell["source"]
            ]
    return notebook_json


def remove_outputs(notebook_json):
    for cell in notebook_json["cells"]:
        if "outputs" in cell:
            cell["outputs"] = []
    return notebook_json


def handle_solutions(notebook_json):
    """
    Take text that looks like
    # BEGIN SOLUTION fitted_dist = ______
    blah blah blah
    # END SOLUTION
    and replace it with
    fitted_dist = ______
    """
    for cell in notebook_json["cells"]:
        if cell["cell_type"] == "code":
            cell["source"] = remove_solution_comments(cell["source"])
    return notebook_json


def remove_solution_comments(cell_source):
    in_solution_block = False
    new_source = []
    for line in cell_source:
        if "# BEGIN SOLUTION" in line:
            new_source.append(line.replace("# BEGIN SOLUTION ", ""))
            in_solution_block = True
        elif "# END SOLUTION" in line:
            assert line.strip() == "# END SOLUTION"
            in_solution_block = False
        elif not in_solution_block:
            new_source.append(line)
    return new_source


def create_skeleton(notebook_json, notebook_path):
    notebook_json = remove_outputs(notebook_json)
    notebook_json = handle_solutions(notebook_json)
    notebook_json = update_links(notebook_json, os.path.basename(notebook_path))
    return notebook_json


def run_notebook(notebook_path):
    assert notebook_path.endswith("_solutions.ipynb")
    with open(notebook_path) as f:
        notebook_json = json.load(f)
    notebook_json = create_skeleton(notebook_json, notebook_path)
    with open(notebook_path.replace("_solutions.ipynb", "_skeleton.ipynb"), "w") as f:
        f.write(json.dumps(notebook_json, indent=1) + "\n")


def run_python(python_path):
    with open(python_path) as f:
        python_code = f.read()
    python_code = python_code.split("\n")
    python_code = remove_solution_comments(python_code)
    python_code = "\n".join(python_code)
    with open(python_path.replace("_solutions.py", "_skeleton.py"), "w") as f:
        f.write(python_code)


def run(path):
    if path.endswith(".ipynb"):
        run_notebook(path)
    elif path.endswith(".py"):
        run_python(path)
    else:
        raise ValueError("Path must be a .ipynb or .py file.")


if __name__ == "__main__":
    import sys

    run(sys.argv[1])



from typing import List

import torch
import tqdm.auto as tqdm

from neurosym.datasets.load_data import DatasetWrapper
from neurosym.dsl.dsl import DSL
from neurosym.examples.near.methods.near_example_trainer import (
    NEARTrainer,
    NEARTrainerConfig,
)
from neurosym.examples.near.models.torch_program_module import TorchProgramModule
from neurosym.examples.near.neural_dsl import PartialProgramNotFoundError
from neurosym.programs.s_expression_render import render_s_expression
from neurosym.search_graph.dsl_search_node import DSLSearchNode
from neurosym.utils.imports import import_pytorch_lightning
from neurosym.utils.logging import log

pl = import_pytorch_lightning()


# callback that updates a progress bar once per epoch
class _ProgressBar(pl.callbacks.Callback):
    def __init__(self, num_epochs, progress_bar):
        self.num_epochs = num_epochs
        self.progress_bar = progress_bar

    def on_train_epoch_end(self, trainer, pl_module):
        self.progress_bar.update(1)
        # set train and val loss in progress bar
        self.progress_bar.set_postfix(
            train_loss=trainer.callback_metrics["train_loss"].item(),
            val_loss=trainer.callback_metrics["val_loss"].item(),
        )


class ValidationCost:
    """
    A class that computes the validation cost of a program using a neural DSL.

    :param neural_dsl: The neural DSL to use.
    :param trainer_cfg: The configuration for the trainer.
    :param datamodule: The data module to use.
    :param error_loss: The loss to return if the program is invalid.
    :param progress_by_epoch: Whether to display progress by epoch.
    :param callbacks: Callbacks to use during training.
    :param kwargs: Additional arguments to pass to the trainer.
    """

    def __init__(
        self,
        *,
        neural_dsl: DSL,
        trainer_cfg: NEARTrainerConfig,
        datamodule: DatasetWrapper,
        error_loss=10000,
        progress_by_epoch=False,
        callbacks: List[pl.callbacks.Callback] = (),
        **kwargs,
    ):
        self.neural_dsl = neural_dsl
        self.trainer_cfg = trainer_cfg
        self.datamodule = datamodule
        self.error_loss = error_loss
        self.kwargs = kwargs
        self.progress_by_epoch = progress_by_epoch
        self.callbacks = list(callbacks)

    def __call__(self, node: DSLSearchNode) -> float:
        trainer, pbar = self._get_trainer_and_pbar(
            label=render_s_expression(node.program)
        )
        try:
            initialized_p = self.neural_dsl.initialize(node.program)
        except PartialProgramNotFoundError:
            return self.error_loss

        model = self.neural_dsl.compute(initialized_p)
        if not isinstance(model, torch.nn.Module):
            del initialized_p
            model = TorchProgramModule(dsl=self.neural_dsl, program=node.program)
        self._fit_trainer(trainer, model, pbar)
        return trainer.callback_metrics["val_loss"].item()

    @staticmethod
    def _duplicate(callbacks):
        """
        Reinitialize all callbacks to avoid sharing state between different validation runs.
        """
        out = []
        for cb in callbacks:
            out.append(
                cb.__class__(
                    **{
                        k: getattr(cb, k)
                        for k in cb.__init__.__code__.co_varnames
                        if hasattr(cb, k)
                    }
                )
            )
        return out

    def _get_trainer_and_pbar(self, label=None):
        callbacks = list(self.callbacks)
        callbacks = self._duplicate(self.callbacks)
        if self.progress_by_epoch:
            log(f"Training {label}")
            pbar = tqdm.tqdm(total=self.trainer_cfg.n_epochs, desc="Training")
            callbacks.append(_ProgressBar(self.trainer_cfg.n_epochs, pbar))
        else:
            pbar = None

        trainer = pl.Trainer(
            max_epochs=self.trainer_cfg.n_epochs,
            devices="auto",
            accelerator="cpu",
            enable_checkpointing=False,
            logger=False,
            **self.kwargs,
            callbacks=callbacks,
        )
        return trainer, pbar

    def _fit_trainer(self, trainer, model, pbar):
        pl_model = NEARTrainer(model, config=self.trainer_cfg)
        trainer.fit(
            pl_model,
            self.datamodule.train_dataloader(),
            self.datamodule.val_dataloader(),
        )
        if self.progress_by_epoch:
            pbar.close()



from neurosym.dsl.dsl_factory import DSLFactory
from neurosym.types.type_string_repr import TypeDefiner


def _increment_and_return(counter):
    counter[0] += 1
    return counter[0]


def _create_mutable_arith_combinators_dsl():
    t = TypeDefiner()
    t.typedef("fn", "(i) -> i")

    dslf = DSLFactory()
    dslf.typedef("fn", "(i) -> i")
    dslf.concrete(
        "x",
        "() -> $fn",
        lambda: lambda t: t,
    )

    dslf.concrete(
        "1",
        "() -> $fn",
        lambda: lambda t: 1,
    )

    dslf.concrete(
        "+",
        "($fn, $fn) -> $fn",
        lambda x, y: lambda t: x(t) + y(t),
    )

    dslf.concrete(
        "*",
        "($fn, $fn) -> $fn",
        lambda x, y: lambda t: x(t) * y(t),
    )

    dslf.concrete(
        "even?",
        "($fn) -> $fn",
        lambda x: lambda t: x(t) % 2 == 0,
    )

    dslf.concrete(
        "ite",
        "($fn, $fn, $fn) -> $fn",
        lambda cond, fx, fy: lambda t: fx(t) if cond(t) else fy(t),
    )

    dslf.parameterized(
        "count",
        "() -> $fn",
        lambda counter: lambda t: _increment_and_return(counter),
        dict(counter=lambda: [0]),
    )
    return dslf.finalize()


# comment is in examples_other.rst
mutable_arith_combinators_dsl = _create_mutable_arith_combinators_dsl()



import json
import unittest

from tutorial.process_tutorial import create_skeleton

from .utils import execute_notebook

validate = """
def validate(t):
    vy = t[:, -1]
    vx = t[:, -2]
    [bounce_loc] = np.where((vy[1:] > 0) & (vy[:-1] < 0))
    if not bounce_loc.size:
        return "no ground bounces"
    # if not (np.abs(vy[bounce_loc]) / np.abs(vy).max() > 0.25).mean() > 0.5:
        # return "ground bounces not at max velocity"
    # if not (np.sign(vx[bounce_loc]) == np.sign(vx[bounce_loc + 1])).mean() > 0.5:
        # return "ground bounces don't mantain vx"
    return "success"


validation = validate(trajectory)
if validation != "success":
    validation = validate(trajectoryb)
print("*" * 80)
print(validation)
"""


class TestBouncingBallExercise(unittest.TestCase):
    def test_bouncing_ball_exercise_output(self):
        result = execute_notebook(
            "tutorial/bouncing_ball_exercise_solutions.ipynb",
            suffix=validate,
            cwd="tutorial",
        )
        *_, stars, validation, _ = result.split("\n")
        self.assertEqual(stars, "*" * 80)
        self.assertEqual(validation, "success")

    def test_bouncing_ball_exercise_skeleton(self):
        self.maxDiff = None
        with open("tutorial/bouncing_ball_exercise_solutions.ipynb") as f:
            solutions = json.load(f)
        with open("tutorial/bouncing_ball_exercise_skeleton.ipynb") as f:
            notebook = json.load(f)

        self.assertEqual(
            create_skeleton(
                solutions, "tutorial/bouncing_ball_exercise_solutions.ipynb"
            ),
            notebook,
        )



"""
Test dsl/sequentual_dsl.py with NEAR search graph.

We conduct the following tests:
- Sanity check: We can find a program.
- BFS: We can find a program with BFS.
- Astar: We can find a program with bounded Astar search.
- Enumerate: We can enumerate all programs of a certain size.
- Full Integration test: Can we run a full iteration of NEAR.
NEAR Integration tests.
"""

import unittest

import pytest

import neurosym as ns
from neurosym.examples import near

from .utils import assertDSLEnumerable


class TestNEARSequentialDSL(unittest.TestCase):
    def test_sequential_dsl_astar(self):
        """
        A minimal implementation of NEAR with a simple DSL.
        search = A-star
        heuristic = validation score after training for N epochs. (pl.Trainer)
        goal = Fully symbolic program. (handled in: search_graph/dsl_search_graph.py)
        test_predicate = score on testing set (pl.Trainer)
        """
        datamodule = ns.datasets.near_data_example(train_seed=0)
        input_dim, output_dim = datamodule.train.get_io_dims()
        original_dsl = near.example_rnn_dsl(input_dim, output_dim)
        trainer_cfg = near.NEARTrainerConfig(
            max_seq_len=100,
            n_epochs=10,
            num_labels=output_dim,
            train_steps=len(datamodule.train),
        )
        t = ns.TypeDefiner(L=input_dim, O=output_dim)
        t.typedef("fL", "{f, $L}")
        t.typedef("fO", "{f, $O}")
        neural_dsl = near.NeuralDSL.from_dsl(
            dsl=original_dsl,
            modules={
                **near.create_modules(
                    "mlp",
                    [t("($fL) -> $fL"), t("($fL) -> $fO")],
                    near.mlp_factory(hidden_size=10),
                ),
                **near.create_modules(
                    "rnn_seq2seq",
                    [t("([$fL]) -> [$fL]"), t("([$fL]) -> [$fO]")],
                    near.rnn_factory_seq2seq(hidden_size=10),
                ),
                **near.create_modules(
                    "rnn_seq2class",
                    [t("([$fL]) -> $fL"), t("([$fL]) -> $fO")],
                    near.rnn_factory_seq2class(hidden_size=10),
                ),
            },
        )

        g = near.near_graph(
            neural_dsl,
            ns.parse_type(
                s="([{f, $L}]) -> [{f, $O}]",
                env=ns.TypeDefiner(L=input_dim, O=output_dim),
            ),
            is_goal=neural_dsl.program_has_no_holes,
        )
        # succeed if this raises StopIteration
        with pytest.raises(StopIteration):
            n_iter = 0
            iterator = ns.search.bounded_astar(
                g,
                near.ValidationCost(
                    neural_dsl=neural_dsl,
                    trainer_cfg=trainer_cfg,
                    datamodule=datamodule,
                ),
                max_depth=3,
            )
            while True:
                print("iteration: ", n_iter)
                n_iter += 1
                node = next(iterator)
                self.assertIsNotNone(node)
                if n_iter > 30:
                    break

    def test_sequential_dsl_enumerate(self):
        """
        Enumerate all programs in dsl upto fixed depth. This test case makes
        sure all DSL combinations upto a fixed depth are valid.
        """
        self.maxDiff = None
        dsl = near.example_rnn_dsl(10, 4)

        assertDSLEnumerable(dsl, "([$fL]) -> [$fO]")



from typing import Callable

from .load_data import DatasetWrapper, numpy_dataset_from_github


def near_data_example(train_seed) -> Callable[[int], DatasetWrapper]:
    """
    Data example from the Near library. Imported from Github.
    Takes a seed and returns a DatasetWrapper object containing the data.
    """

    return numpy_dataset_from_github(
        "https://github.com/trishullab/near/tree/master/near_code/data/example",
        "train_ex_data.npy",
        "train_ex_labels.npy",
        "test_ex_data.npy",
        "test_ex_labels.npy",
    )(train_seed)



from collections import OrderedDict
from dataclasses import dataclass
from typing import List, Tuple

import torch
from torch import nn

from .base import BaseConfig


@dataclass
class MLPConfig(BaseConfig):
    """
    Represents the configuration of an MLP module.

    :param input_size: The size of the input.
    :param hidden_size: The size of the hidden layer.
    :param output_size: The size of the output.
    :param dropout: Dropout rate to apply to the hidden layer.
    :param bias: Whether to use bias in the linear layers.
    :param nonlinearity: Nonlinearity to use in the hidden layer.
    :param loss: Loss function to use for training the model.
    """

    input_size: int
    hidden_size: int
    output_size: int
    dropout: float = 0.0
    bias: bool = True
    nonlinearity: str = "LeakyReLU"
    loss: str = "MSELoss"


class MLP(nn.Sequential):
    """
    Simple MLP module.

    :param config: Configuration for the MLP.
    """

    def __init__(self, config: MLPConfig):
        self.config = config
        nonlinearity = getattr(nn, config.nonlinearity)
        model_dict = OrderedDict(
            fc1=nn.Linear(config.input_size, config.hidden_size, bias=config.bias),
            act1=nonlinearity(),
            dp1=nn.Dropout(p=config.dropout),
            fc3=nn.Linear(config.hidden_size, config.output_size, bias=config.bias),
        )
        super().__init__(model_dict)

        @torch.no_grad()
        def init_weights(m):
            if isinstance(m, nn.Linear):
                m.weight = nn.init.xavier_normal_(
                    m.weight, gain=torch.nn.init.calculate_gain("leaky_relu")
                )

        self.apply(init_weights)


def mlp_factory(
    hidden_size: int,
    dropout: float = 0.0,
    bias: bool = True,
    nonlinearity: str = "LeakyReLU",
    loss: str = "MSELoss",
):
    """
    Allows instantiating an MLP module with a given input and output size.

    :param hidden_size: Number of hidden units in the MLP.
    :param dropout: Dropout rate to apply to the hidden layer.
    :param bias: Whether to use bias in the linear layers.
    :param nonlinearity: Nonlinearity to use in the hidden layer.
    :param loss: Loss function to use for training the model.
    """

    def construct_model(input_shape: List[Tuple[int]], output_shape: Tuple[int]):
        assert len(input_shape) == 1, "MLP takes a single input only."
        input_size = input_shape[0][-1]
        output_size = output_shape[-1]
        cfg = MLPConfig(
            model_name="mlp",
            input_size=input_size,
            output_size=output_size,
            hidden_size=hidden_size,
            dropout=dropout,
            bias=bias,
            nonlinearity=nonlinearity,
            loss=loss,
        )
        return MLP(cfg)

    return construct_model



def fold_torch(func, lst):  # noqa: E741
    """
    Runs a fold operation on a list of tensors, handling the batch dimension.

    :param func: ``((N, *C), (N, *C)) -> (N, *C)``
    :param lst: ``(N, L, *C)``
    :returns: ``(N, *C)``
    """
    result = lst[:, 0, :]
    for i in range(1, lst.shape[1]):
        result = func(result, lst[:, i, :])
    return result


def map_torch(func, lst):  # noqa: E741
    """

    Runs a map operation on a list of tensors, handling the batch dimension.

    :param func: ``(N, *C1) -> (N, *C2)``
    :param lst: ``(N, L, *C1)``
    :returns: ``(N, L, *C2)``
    """
    original_shape = lst.shape
    reshaped = lst.reshape(lst.shape[0] * lst.shape[1], *lst.shape[2:])
    result = func(reshaped)
    return result.reshape(original_shape[0], original_shape[1], *result.shape[1:])



import torch


def ite_torch(condition, if_true, if_else):
    """
    Computes a differentiable if-then-else operation.

    :param condition: A function that computes the condition. This is
        expected to return a tensor of shape ``(N, *C1)``, and the output
        will be broadcast to the shape of the other arguments. The
        condition will have a sigmoid applied to it to put it in the
        range [0, 1], so positive values will be interpreted as true,
        and negative values as false.

    :param if_true: A function that computes the value if the condition
        is true. This is expected to return a tensor of shape ``(N, *C2)``.

    :param if_else: A function that computes the value if the condition
        is false. This is expected to return a tensor of shape ``(N, *C2)``.
    """

    def _ite(*args):
        cond = torch.sigmoid(condition(*args))
        true_val = if_true(*args)
        false_val = if_else(*args)
        if len(cond.shape) == len(true_val.shape) - 1:
            cond = cond.unsqueeze(1)
        cond = cond.expand_as(true_val)
        return cond * true_val + (1 - cond) * false_val

    return _ite
